"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
from typing import Any, Optional, Callable, Coroutine, TYPE_CHECKING

import aiohttp

from .enums import ChatCmd
from .gateway import ChzzkWebSocket, ReconnectWebsocket
from .http import ChzzkChatSession
from .state import ConnectionState
from ..client import Client
from ..error import LoginRequired
from ..http import ChzzkAPISession

if TYPE_CHECKING:
    from .access_token import AccessToken
    from .message import ChatMessage
    from .recent_chat import RecentChat

_log = logging.getLogger(__name__)


class ChatClient(Client):
    def __init__(
        self,
        channel_id: str,
        authorization_key: Optional[str] = None,
        session_key: Optional[str] = None,
        chat_channel_id: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        super().__init__(
            loop=loop, authorization_key=authorization_key, session_key=session_key
        )

        self.chat_channel_id: str = chat_channel_id
        self.channel_id: str = channel_id
        self.access_token: Optional[AccessToken] = None
        self.user_id: Optional[str] = None

        self.ws_session = aiohttp.ClientSession(loop=self.loop)

        self._listeners: dict[str, list[tuple[asyncio.Future, Callable[..., bool]]]] = (
            dict()
        )
        self._extra_event: dict[str, list[Callable[..., Coroutine[Any, Any, Any]]]] = (
            dict()
        )

        self._ready = asyncio.Event()

        handler = {ChatCmd.CONNECTED: self._ready.set}
        self._connection = ConnectionState(dispatch=self.dispatch, handler=handler)
        self._gateway: Optional[ChzzkWebSocket] = None

    def _session_initial_set(self):
        self._api_session = ChzzkAPISession(loop=self.loop)
        self._game_session = ChzzkChatSession(loop=self.loop)

    @property
    def is_connected(self) -> bool:
        return self._ready.is_set()

    def run(self, authorization_key: str = None, session_key: str = None) -> None:
        wrapper = self.start(authorization_key, session_key)
        try:
            self.loop.run_until_complete(wrapper)
        except KeyboardInterrupt:
            return

    async def start(self, authorization_key: str = None, session_key: str = None):
        try:
            if authorization_key is not None and session_key is not None:
                self.login(authorization_key=authorization_key, session_key=session_key)
            await self.connect()
        finally:
            await self.close()

    async def connect(self) -> None:
        if self.chat_channel_id is None:
            status = await self.live_status(channel_id=self.channel_id)
            self.chat_channel_id = status.chat_channel_id

        if self._game_session.has_login:
            user = await self.user()
            self.user_id = user.user_id_hash

        if self.access_token is None:
            await self._generate_access_token()

        await self.polling()

    async def close(self):
        self._ready.clear()

        if self._gateway is not None:
            await self._gateway.socket.close()
        await self.ws_session.close()
        await super().close()

    async def polling(self) -> None:
        session_id: Optional[str] = None
        while not self.is_closed:
            try:
                self._gateway = await ChzzkWebSocket.from_client(
                    self, self._connection, session_id=session_id
                )

                # Initial Connection
                if session_id is None:
                    await self._gateway.send_open(
                        access_token=self.access_token.access_token,
                        chat_channel_id=self.chat_channel_id,
                        mode="READ" if self.user_id is None else "SEND",
                        user_id=self.user_id,
                    )
                    session_id = self._gateway.session_id

                while True:
                    await self._gateway.poll_event()
            except ReconnectWebsocket:
                self.dispatch("disconnect")
                continue

    # Event Handler
    async def wait_until_connected(self) -> None:
        await self._ready.wait()

    def wait_for(
        self,
        event: str,
        check: Optional[Callable[..., bool]] = None,
        timeout: Optional[float] = None,
    ):
        future = self.loop.create_future()

        if check is None:

            def _check(*_):
                return True

            check = _check
        event_name = event.lower()

        if event_name not in self._listeners.keys():
            self._listeners[event_name] = list()
        self._listeners[event_name].append((future, check))
        return asyncio.wait_for(future, timeout=timeout)

    def event(
        self, coro: Callable[..., Coroutine[Any, Any, Any]]
    ) -> Callable[..., Coroutine[Any, Any, Any]]:
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("function must be a coroutine.")

        event_name = coro.__name__
        if event_name not in self._listeners.keys():
            self._extra_event[event_name] = list()
        self._extra_event[event_name].append(coro)
        return coro

    def dispatch(self, event: str, *args: Any, **kwargs) -> None:
        _log.debug("Dispatching event %s", event)
        method = "on_" + event

        # wait-for listeners
        if event in self._listeners.keys():
            listeners = self._listeners[event]
            _new_listeners = []

            for index, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    continue

                try:
                    result = condition(*args, **kwargs)
                except Exception as e:
                    future.set_exception(e)
                    continue
                if result:
                    match len(args):
                        case 0:
                            future.set_result(None)
                        case 1:
                            future.set_result(args[0])
                        case _:
                            future.set_result(args)

                _new_listeners.append((future, condition))
            self._listeners[event] = _new_listeners

        # event-listener
        if method not in self._extra_event.keys():
            return

        for coroutine_function in self._extra_event[method]:
            self._schedule_event(coroutine_function, method, *args, **kwargs)

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            try:
                _log.exception("Ignoring exception in %s", event_name)
                self.dispatch("error", exc, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    def _schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> asyncio.Task:
        wrapped = self._run_event(coro, event_name, *args, **kwargs)
        # Schedules the task
        return self.loop.create_task(wrapped, name=f"chzzk.py: {event_name}")

    # API Method
    async def _generate_access_token(self) -> AccessToken:
        res = await self._game_session.chat_access_token(
            channel_id=self.chat_channel_id
        )
        self.access_token = res.content
        return self.access_token

    # Chat Method
    async def send_chat(self, message: str) -> None:
        if not self.is_connected:
            raise RuntimeError("Not connected to server. Please connect first.")

        if not self.user_id:
            raise LoginRequired()

        await self._gateway.send_chat(message, self.chat_channel_id)

    async def request_recent_chat(self, count: int = 50):
        if not self.is_connected:
            raise RuntimeError("Not connected to server. Please connect first.")

        await self._gateway.request_recent_chat(count, self.chat_channel_id)

    async def history(self, count: int = 50) -> list[ChatMessage]:
        await self.request_recent_chat(count)
        recent_chat: RecentChat = await self.wait_for(
            "recent_chat", lambda x: len(recent_chat.message_list) <= count
        )
        return recent_chat.message_list



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import functools
import logging
from typing import Annotated, Optional

from ahttp_client import Session, get, Path
from ahttp_client.extension import get_pydantic_response_model
from ahttp_client.request import RequestCore

from .base_model import ChzzkModel, Content
from .error import LoginRequired
from .live import LiveStatus, LiveDetail
from .user import User

_log = logging.getLogger(__name__)


class ChzzkSession(Session):
    def __init__(self, base_url: str, loop: Optional[asyncio.AbstractEventLoop] = None):
        super().__init__(base_url=base_url, loop=loop)

        self._authorization_key = None
        self._session_key = None

    def login(self, authorization_key: str, session_key: str):
        self._authorization_key = authorization_key
        self._session_key = session_key

    @property
    def has_login(self) -> bool:
        return self._authorization_key is not None and self._session_key is not None

    @staticmethod
    def configuration(login_able: bool = False, login_required: bool = False):
        def decorator(func):
            func.__login_able__ = login_able
            func.__login_required__ = login_required
            return func

        return decorator

    async def before_request(
        self, request: RequestCore, path: str
    ) -> tuple[RequestCore, str]:
        _log.debug(f"Path({path}) was called.")

        # Authorization
        if getattr(request.func, "__login_able__", False):
            if self.has_login:
                if "Cookie" not in request.headers.keys():
                    request.headers["Cookie"] = ""
                request.headers["Cookie"] += self._token
            elif getattr(request.func, "__login_required__", False):
                raise LoginRequired()
        return request, path

    @property
    def _token(self) -> str:
        return f"NID_SES={self._session_key}; NID_AUT={self._authorization_key}"


class ChzzkAPISession(ChzzkSession):
    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        super().__init__(base_url="https://api.chzzk.naver.com", loop=loop)

    @get_pydantic_response_model()
    @get("/polling/v2/channels/{channel_id}/live-status", directly_response=True)
    async def live_status(
        self, channel_id: Annotated[str, Path]
    ) -> Content[LiveStatus]:
        pass

    @get_pydantic_response_model()
    @get("/service/v2/channels/{channel_id}/live-detail", directly_response=True)
    async def live_detail(
        self, channel_id: Annotated[str, Path]
    ) -> Content[LiveDetail]:
        pass


class NaverGameAPISession(ChzzkSession):
    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        super().__init__(base_url="https://comm-api.game.naver.com", loop=loop)

    @get_pydantic_response_model()
    @get("/nng_main/v1/user/getUserStatus", directly_response=True)
    @ChzzkSession.configuration(login_able=True, login_required=True)
    async def user(self) -> Content[User]:
        pass



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from enum import IntEnum, Enum
from typing import TypeVar, Any

E = TypeVar("E", bound="Enum")


class ChatCmd(IntEnum):
    PING = 0
    PONG = 10000
    CONNECT = 100
    CONNECTED = 10100
    REQUEST_RECENT_CHAT = 5101
    RECENT_CHAT = 15101
    EVENT = 93006
    CHAT = 93101
    SPECIAL_CHAT = 93102  # Donation / System Message
    KICK = 94005
    BLOCK = 94006
    BLIND = 94008
    NOTICE = 94010
    PENALTY = 94015
    SEND_CHAT = 3101


class ChatType(IntEnum):
    TEXT = 1
    IMAGE = 2
    STICKER = 3
    VIDEO = 4
    RICH = 5
    DONATION = 10
    SYSTEM_MESSAGE = 30
    OPEN = 121


class UserRole(Enum):
    common_user = "common_user"
    streamer = "streamer"
    chat_manager = "streaming_chat_manager"
    channel_manager = "streaming_channel_manager"
    manager = "manager"


def get_enum(cls: type[E], val: Any) -> E:
    enum_val = [i for i in cls if i.value == val]
    if len(enum_val) == 0:
        return val
    return enum_val[0]



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import NamedTuple, Literal, Optional

from .client import Client
from .error import *
from .live import LiveStatus, LiveDetail, LivePollingStatus
from .user import User

# Extension Package
try:
    from .chat import *
except ModuleNotFoundError:
    pass


__title__ = "chzzkpy"
__author__ = "gunyu1019"
__license__ = "MIT"
__copyright__ = "Copyright 2024-present gunyu1019"
__version__ = "0.0.4-alpha1"  # version_info.to_string()


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    release_level: Optional[Literal["alpha", "beta", "candidate", "final"]]
    serial: int

    def to_string(self) -> str:
        _version_info = f"{self.major}.{self.minor}.{self.micro}"
        if self.release_level is not None:
            _version_info += "-{}".format(self.release_level) + str(self.serial)
        return _version_info


version_info: VersionInfo = VersionInfo(
    major=0, minor=0, micro=4, release_level="alpha", serial=1
)



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio

from typing import Optional, TYPE_CHECKING
from .http import ChzzkAPISession, NaverGameAPISession
from .live import LiveStatus, LiveDetail
from .user import User

if TYPE_CHECKING:
    from types import TracebackType
    from typing_extensions import Self


class Client:
    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        authorization_key: Optional[str] = None,
        session_key: Optional[str] = None,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self._closed = False
        self._api_session = None
        self._game_session = None

        if authorization_key is not None and session_key is not None:
            self.login(authorization_key, session_key)

        self._session_initial_set()

    def _session_initial_set(self):
        self._api_session = ChzzkAPISession(loop=self.loop)
        self._game_session = NaverGameAPISession(loop=self.loop)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if not self.is_closed:
            await self.close()

    @property
    def is_closed(self) -> bool:
        return self._closed

    async def close(self):
        self._closed = True
        await self._api_session.close()
        await self._game_session.close()
        return

    def login(self, authorization_key: str, session_key: str):
        self._api_session.login(authorization_key, session_key)
        self._game_session.login(authorization_key, session_key)

    async def live_status(self, channel_id: str) -> Optional[LiveStatus]:
        res = await self._api_session.live_status(channel_id=channel_id)
        return res.content

    async def live_detail(self, channel_id: str) -> Optional[LiveDetail]:
        res = await self._api_session.live_detail(channel_id=channel_id)
        return res.content

    async def user(self) -> User:
        res = await self._game_session.user()
        return res.content



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from typing import Optional, Literal, TypeVar, Generic, Any
from pydantic import AliasChoices, Field, Json

from .enums import ChatType
from .profile import Profile
from ..base_model import ChzzkModel

E = TypeVar("E", bound="ExtraBase")


class ExtraBase(ChzzkModel):
    pass


class Extra(ExtraBase):
    chat_type: str
    emojis: Any
    os_type: Literal["PC", "AOS", "IOS"]
    streaming_channel_id: str


class Message(ChzzkModel, Generic[E]):
    service_id: str = Field(validation_alias=AliasChoices("serviceId", "svcid"))
    channel_id: str = Field(validation_alias=AliasChoices("channelId", "cid"))
    user_id: str = Field(validation_alias=AliasChoices("uid", "userId"))

    profile: Optional[Json[Profile]]
    content: str = Field(validation_alias=AliasChoices("msg", "content"))
    type: ChatType = Field(
        validation_alias=AliasChoices("msgTypeCode", "messageTypeCode")
    )
    extras: Optional[Json[E]]

    created_time: datetime.datetime = Field(
        validation_alias=AliasChoices("ctime", "createTime")
    )
    updated_time: Optional[datetime.datetime] = Field(
        default=None, validation_alias=AliasChoices("utime", "updateTime")
    )
    time: datetime.datetime = Field(
        validation_alias=AliasChoices("msgTime", "messageTime")
    )


class MessageDetail(Message[E], Generic[E]):
    member_count: int = Field(validation_alias=AliasChoices("mbrCnt", "memberCount"))
    message_status: Optional[str] = Field(
        validation_alias=AliasChoices("msgStatusType", "messageStatusType")
    )

    # message_tid: ???
    # session: bool

    @property
    def is_blind(self) -> bool:
        return self.message_status == "BLIND"


class ChatMessage(MessageDetail[Extra]):
    pass


class NoticeExtra(Extra):
    register_profile: Profile


class NoticeMessage(Message[NoticeExtra]):
    pass


class DonationRank(ChzzkModel):
    user_id_hash: str
    nickname: str = Field(validation_alias=AliasChoices("nickname", "nickName"))
    verified_mark: bool
    donation_amount: int
    ranking: int


class DonationExtra(ExtraBase):
    is_anonymous: bool = True
    pay_type: str
    pay_amount: int = 0
    donation_type: str
    weekly_rank_list: Optional[list[DonationRank]] = Field(default_factory=list)
    donation_user_weekly_rank: Optional[DonationRank] = None


class DonationMessage(MessageDetail[DonationExtra]):
    pass


class SystemExtraParameter(ChzzkModel):
    register_nickname: str
    target_nickname: str
    register_chat_profile: Json[Profile] = Field(alias="registerChatProfileJson")
    target_profile: Json[Profile] = Field(alias="targetChatProfileJson")


class SystemExtra(ExtraBase):
    description: str
    style_type: int
    visible_roles: list[str]
    params: Optional[SystemExtraParameter] = None


class SystemMessage(MessageDetail[SystemExtra]):
    pass



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Any, Optional
from pydantic import computed_field, Field, PrivateAttr

from .enums import UserRole
from ..base_model import ChzzkModel


class Badge(ChzzkModel):
    name: Optional[str] = None
    image_url: Optional[str] = None


class StreamingProperty(ChzzkModel):
    _following_dt: Optional[dict[str, str]] = PrivateAttr(default=None)
    _real_time_donation_ranking_dt: Optional[dict[str, str]] = PrivateAttr(default=None)

    def __init__(self, **kwargs):
        super(ChzzkModel, self).__init__(**kwargs)
        self._following_dt = kwargs.pop("following", None)
        self._real_time_donation_ranking_dt = kwargs.pop(
            "realTimeDonationRanking", None
        )

    @computed_field
    @property
    def following_date(self) -> Optional[str]:
        if self._following_dt is None:
            return
        return self._following_dt["followDate"]

    @computed_field
    @property
    def donation_ranking_badge(self) -> Optional[Badge]:
        if (
            self._real_time_donation_ranking_dt is None
            or "badge" not in self._real_time_donation_ranking_dt.keys()
        ):
            return
        return Badge.model_validate_json(self._real_time_donation_ranking_dt["badge"])


class ActivityBadge(Badge):
    badge_no: int
    badge_id: str
    description: Optional[str] = None
    activated: bool


class Profile(ChzzkModel):
    activity_badges: list[Any]
    user_id_hash: str
    user_role: Optional[UserRole] = Field(alias="userRoleCode", default=None)
    nickname: str
    profile_image_url: Optional[str]
    _badge: Optional[dict[str, str]] = PrivateAttr(default=None)
    _title: Optional[dict[str, str]] = PrivateAttr(default=None)
    streaming_property: Optional[StreamingProperty] = None
    activity_badges: list[ActivityBadge] = Field(default_factory=list)
    verified_mark: bool = False

    def __init__(self, **kwargs):
        super(ChzzkModel, self).__init__(**kwargs)
        self._badge = kwargs.pop("badge", None)
        self._title = kwargs.pop("title", None)

    @computed_field
    @property
    def color(self) -> Optional[str]:
        if self._title is None:
            return
        return self._title["color"]

    @computed_field
    @property
    def badge(self) -> Optional[Badge]:
        if self._badge is None and self._title is None:
            return
        _badge = self._badge or dict()
        _title = self._title or dict()
        return Badge(
            name=_title.get("name", None), image_url=_badge.get("imageUrl", None)
        )



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import json
import logging
import time
from typing import Any, Callable, Optional, Literal, TYPE_CHECKING

import aiohttp

from .enums import ChatCmd, get_enum, ChatType
from .error import ConnectionClosed, WebSocketClosure, ReconnectWebsocket

if TYPE_CHECKING:
    from typing_extensions import Self

    from .chat_client import ChatClient
    from .state import ConnectionState

_log = logging.getLogger(__name__)


class ChzzkWebSocket:
    def __init__(
        self,
        socket: aiohttp.ClientWebSocketResponse,
        loop: asyncio.AbstractEventLoop,
    ):
        self.socket: aiohttp.ClientWebSocketResponse = socket
        self.loop: asyncio.AbstractEventLoop = loop
        self.session_id: Optional[str] = None

        self._max_timeout: float = 60.0

        self._event_hook: dict[ChatCmd, Optional[Callable[..., Any]]] = {
            key: None for key in list(ChatCmd)
        }

    def set_hook(self, cmd: ChatCmd, coro_func: Callable[..., Any]):
        self._event_hook[cmd] = coro_func

    def remove_hook(self, cmd: ChatCmd):
        self._event_hook[cmd] = None

    @classmethod
    async def new_session(
        cls,
        loop: asyncio.AbstractEventLoop,
        session: aiohttp.ClientSession,
        channel_id: str,
        session_id: Optional[str] = None,
    ) -> Self:
        server_id = abs(sum([ord(x) for x in channel_id])) % 9 + 1
        url = f"wss://kr-ss{server_id}.chat.naver.com/chat"
        socket: aiohttp.ClientWebSocketResponse = await session.ws_connect(url)

        websocket = cls(socket, loop)
        websocket.session_id = session_id
        return websocket

    @classmethod
    async def from_client(
        cls,
        client: "ChatClient",
        state: "ConnectionState",
        session_id: Optional[str] = None,
    ) -> Self:
        websocket = await cls.new_session(
            loop=client.loop,
            session=client.ws_session,
            channel_id=client.chat_channel_id,
            session_id=session_id,
        )
        for cmd, parsing_func in state.parsers.items():
            if parsing_func is None:
                continue
            websocket.set_hook(cmd, parsing_func)
        return websocket

    @property
    def default_body(self) -> dict[str, str]:
        return {"svcid": "game", "ver": "2"}

    def _can_handle_close(self, code: Optional[int] = None) -> bool:
        if code is None:
            code = self.socket.close_code
        return code != 1000

    async def poll_event(self):
        try:
            msg = await self.socket.receive(timeout=59.0)
            if msg.type is aiohttp.WSMsgType.TEXT:
                data = msg.json()
                await self.received_message(data)
            elif msg.type is aiohttp.WSMsgType.ERROR:
                _log.debug("Received error %s", msg)
                raise WebSocketClosure
            elif msg.type in (
                aiohttp.WSMsgType.CLOSED,
                aiohttp.WSMsgType.CLOSING,
                aiohttp.WSMsgType.CLOSE,
            ):
                _log.debug("Received %s", msg)
                raise WebSocketClosure
        except asyncio.TimeoutError:
            await self.send_ping()
            _log.debug("Timeout receiving packet. Send to ping for keep-alive")
        except WebSocketClosure:
            code = self.socket.close_code
            if self._can_handle_close(code):
                raise ReconnectWebsocket()
            else:
                raise ConnectionClosed(self.socket, code)

    async def received_message(self, data: dict[str, Any]) -> None:
        cmd: int = data["cmd"]
        body = data.get("bdy")

        # service_id = data['svcid']
        # channel_id = data['cid']
        # version = data['ver']
        # tid: ? = data.get('tid')

        cmd_type = get_enum(ChatCmd, cmd)
        _log.debug("Received Message: %s", data)

        if cmd_type == ChatCmd.CONNECTED:
            self.session_id = body["sid"]
        elif cmd_type == ChatCmd.PING:
            await self.send_pong()
            return

        func = self._event_hook.get(cmd_type)
        if func is not None:
            func(body)

    async def send(self, data: str) -> None:
        _log.debug("Sending data: %s", data)
        await self.socket.send_str(data)

    async def send_json(self, data: dict[str, Any]) -> None:
        _log.debug("Sending JSON: %s", json.dumps(data))
        await self.socket.send_json(data)

    async def send_pong(self):
        await self.send_json({"cmd": ChatCmd.PONG, "ver": 2})

    async def send_ping(self):
        await self.send_json({"cmd": ChatCmd.PING, "ver": 2})

    async def send_open(
        self,
        access_token: str,
        chat_channel_id: str,
        mode: Literal["SEND", "READ"],
        user_id: Optional[str] = None,
    ):
        data: dict[str, Any] = {
            "bdy": {
                "accTkn": access_token,
                "auth": mode,
                "devType": 2001,
                "uid": user_id,
            },
            "cid": chat_channel_id,
            "cmd": ChatCmd.CONNECT,
            "tid": 1,
        }
        data.update(self.default_body)
        await self.send_json(data)

    async def send_chat(self, message: str, chat_channel_id: str):
        extra: dict[str, Any] = {
            "chatType": "STREAMING",
            "emojis": "",
            "osType": "PC",
            "streamingChannelId": chat_channel_id,
        }

        data: dict[str, Any] = {
            "bdy": {
                "extras": json.dumps(extra),
                "msg": message,
                "msgTime": int(time.time() * 1000),
                "msgTypeCode": ChatType.TEXT,
            },
            "retry": False,
            "cmd": ChatCmd.SEND_CHAT,
            "sid": self.session_id,
            "cid": chat_channel_id,
            "tid": 3,
        }
        data.update(self.default_body)
        await self.send_json(data)

    async def request_recent_chat(self, count: int, chat_channel_id: str):
        data: dict[str, Any] = {
            "bdy": {"recentMessageCount": count},
            "cmd": ChatCmd.REQUEST_RECENT_CHAT,
            "sid": self.session_id,
            "cid": chat_channel_id,
            "tid": 2,
        }
        data.update(self.default_body)
        await self.send_json(data)



"""MIT License

Copyright (c) 2024 gunyu1019

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from typing import Literal, Any, Optional
from pydantic import Field, Json
from .base_model import ChzzkModel


class LivePollingStatus(ChzzkModel):
    status: str
    is_publishing: bool
    playable_status: str
    traffic_throttling: int
    call_period_millisecond: int = Field(alias="callPeriodMilliSecond")


class LiveStatus(ChzzkModel):
    live_title: str
    status: Literal["OPEN", "CLOSE"]
    concurrent_user_count: int
    accumulate_count: int
    paid_promotion: bool
    adult: bool
    chat_channel_id: str
    category_type: Optional[str]
    live_category: Optional[str]
    live_category_value: str
    live_polling_status: Json[LivePollingStatus] = Field(alias="livePollingStatusJson")
    fault_status: Any  # typing: ???
    user_adult_status: str
    chat_active: bool
    chat_available_group: str
    chat_available_condition: str
    min_follower_minute: int


class LiveDetail(ChzzkModel):
    live_id: int
    live_title: str
    status: Literal["OPEN", "CLOSE"]
    live_image_url: str
    default_thumbnail_image_url: Optional[str]
    concurrent_user_count: int
    accumulate_count: int
    open_date: datetime.datetime
    close_date: datetime.datetime
    adult: bool
    chat_channel_id: str
    category_type: Optional[str]
    live_category: str
    live_category_value: str
    chat_active: bool
    chat_available_group: str
    chat_available_condition: str
    paid_promotion: bool
    min_follower_minute: int
    user_adult_status: Optional[str]

    # live_playback: Json[LivePlayback] = Field(alias="livePlaybackJson") WIP
    live_polling_status: Json[LivePollingStatus] = Field(alias="livePollingStatusJson")



from ahttp_client import get, Query
from ahttp_client.extension import get_pydantic_response_model
from typing import Annotated

from .access_token import AccessToken
from ..base_model import Content
from ..http import ChzzkSession, NaverGameAPISession


class ChzzkChatSession(NaverGameAPISession):
    @get_pydantic_response_model()
    @get("/nng_main/v1/chats/access-token", directly_response=True)
    @ChzzkSession.configuration(login_able=True)
    @Query.default_query("chatType", "STREAMING")
    async def chat_access_token(
        self, channel_id: Annotated[str, Query.to_camel()]
    ) -> Content[AccessToken]:
        pass
